from noise_contrast import noise_contrast, NoiseContraster
from w2v import word2vec, sigmoid, row_dot, Word2VecEmbedder
from token_map import TokenMap
from minibatcher import (
	Minibatcher, Word2VecMinibatcher, TokenChooser, default_parse
)
from counter_sampler import (
	CounterSampler, CounterSamplerException, MultinomialSampler
)
from unigram_dictionary import UnigramDictionary



from unigram_dictionary import UnigramDictionary
from collections import Counter, defaultdict
from token_map import TokenMap, SILENT, ERROR
import time
from unittest import main, TestCase
from theano import tensor as T, function
import numpy as np
from w2v import Word2VecEmbedder, word2vec
from noise_contrast import noise_contrast, NoiseContraster
from minibatcher import Word2VecMinibatcher, TokenChooser, default_parse
from counter_sampler import MultinomialSampler, CounterSampler
from lasagne.init import Normal


def sigma(a):
	return 1/(1+np.exp(-a))
usigma = np.vectorize(sigma)


class TestUnigramDictionary(TestCase):
	'''
	Tests that UnigramDictionary properly represents the corpus 
	statistics, and that the pruning function works as expected.
	'''

	TOKENS = ['apple', 'pear', 'banana', 'orange']

	# Make a toy corpus with specific token frequencies.
	FREQUENCIES = {
		'apple':4, 'banana':8, 'orange':6, 
		'pineapple':3, 'grapefruit':9
	}
	CORPUS = list(Counter(FREQUENCIES).elements())


	# TODO test that we actually produce the samples according to the 
	# frequency distribution of the corpus
	def test_sampling(self):
		'''
		Test basic function of assigning counts, and then sampling from
		The distribution implied by those counts.
		'''
		unigram_dictionary = UnigramDictionary()
		unigram_dictionary.update(self.FREQUENCIES)

		# Test asking for a single sample (where no shape tuple supplied)
		single_sample = unigram_dictionary.sample()
		self.assertTrue(type(single_sample) is np.int64)

		# Test asking for an array of samples (by passing a shape tuple)
		shape = (2,3,5)
		array_sample = unigram_dictionary.sample(shape)
		self.assertTrue(type(array_sample) is np.ndarray)
		self.assertTrue(array_sample.shape == shape)


	def test_counter_sampler_statistics(self):
		'''
		This tests that the UnigramDictionary really does produce results 
		whose statistics match those requested by the counts vector
		'''
		# Seed numpy's random function to make the test reproducible
		np.random.seed(1)

		# Make a sampler with probabilities proportional to counts
		unigram_dictionary = UnigramDictionary()
		unigram_dictionary.update(self.CORPUS)

		# Draw one hundred thousand samples, then total up the fraction of
		# each outcome obseved
		counter = Counter(unigram_dictionary.sample((100000,)))

		# Make a list of the expected fractions by which each outcome
		# should be observed, in the limit of infinite sample
		total_in_expected = float(len(self.CORPUS))

		tolerance = 0.002
		for idx, found_freq in counter.iteritems():
			found_frac = found_freq / 100000.0
			token = unigram_dictionary.get_token(idx)
			expected_frac = self.FREQUENCIES[token] / total_in_expected
			self.assertTrue(abs(found_frac - expected_frac) < tolerance)


	def test_unigram_dictionary_token_map(self):

		unigram_dictionary = UnigramDictionary(on_unk=SILENT)

		for idx, fruit in enumerate(self.TOKENS):
			# Ensure that ids are assigned in an auto-incrementing way
			# starting from 1 (0 is reserved for the UNK token)
			self.assertEqual(unigram_dictionary.add(fruit), idx+1)

		for idx, fruit in enumerate(self.TOKENS):
			# Ensure that idxs are stable and retrievable with 
			# UnigramDictionary.get_id()
			self.assertEqual(unigram_dictionary.get_id(fruit), idx+1)

			# Ensure that we can look up the token using the id
			self.assertEqual(unigram_dictionary.get_token(idx+1), fruit)

		# Ensure the unigram_dictionary knows its own length
		self.assertEqual(len(unigram_dictionary), len(self.TOKENS)+1)

		# Asking for ids of non-existent tokens returns the UNK token_id
		self.assertEqual(unigram_dictionary.get_id('no-exist'), 0)

		# Asking for the token at 0 returns 'UNK'
		self.assertEqual(unigram_dictionary.get_token(0), 'UNK')

		# Asking for token at non-existent idx raises IndexError
		with self.assertRaises(IndexError):
			unigram_dictionary.get_token(99)


	def test_raise_error_on_unk(self):
		'''
		If the unigram_dictionary is constructed passing 
			on_unk=UnigramDictionary.ERROR
		then calling get_id() or get_ids() will throw a KeyError if one
		of the supplied tokens isn't in the unigram_dictionary.  
		(Normally it would return 0, which is a token id reserved for 
		'UNK' -- any unknown token).
		'''

		unigram_dictionary = UnigramDictionary(on_unk=ERROR)
		unigram_dictionary.update(self.TOKENS)

		with self.assertRaises(KeyError):
			unigram_dictionary.get_id('no-exist')

		with self.assertRaises(KeyError):
			unigram_dictionary.get_ids(['apple', 'no-exist'])



	def test_plural_functions(self):

		unigram_dictionary = UnigramDictionary(on_unk=SILENT)

		# In these assertions, we offset the expected list of ids by 1 
		# because the 0th id in unigram_dictionary is reserved for the UNK
		# token

		# Ensure that update works
		ids = unigram_dictionary.update(self.TOKENS)
		self.assertEqual(ids, range(1, len(self.TOKENS)+1))

		# Ensure that get_ids works
		self.assertEqual(
			unigram_dictionary.get_ids(self.TOKENS),
			range(1, len(self.TOKENS)+1)
		)

		# Ensure that get_tokens works
		self.assertEqual(
			unigram_dictionary.get_tokens(range(1, len(self.TOKENS)+1)),
			self.TOKENS
		)

		# Asking for ids of non-existent tokens raises KeyError
		self.assertEqual(
			unigram_dictionary.get_ids(['apple', 'no-exist']),
			[self.TOKENS.index('apple')+1, 0]
		)

		# Asking for token at 0 returns the 'UNK' token
		self.assertEqual(
			unigram_dictionary.get_tokens([3,0]),
			[self.TOKENS[3-1], 'UNK']
		)

		# Asking for token at non-existent idx raises IndexError
		with self.assertRaises(IndexError):
			unigram_dictionary.get_tokens([1,99])


	def test_save_load(self):

		unigram_dictionary = UnigramDictionary(on_unk=SILENT)
		unigram_dictionary.update(self.CORPUS)
		unigram_dictionary.save('test-data/test-unigram-dictionary')

		unigram_dictionary_copy = UnigramDictionary(on_unk=SILENT)
		unigram_dictionary_copy.load('test-data/test-unigram-dictionary')

		# Test that the mapping from tokens to ids is unchanged
		for token in self.FREQUENCIES:
			self.assertEqual(
				unigram_dictionary.get_id(token),
				unigram_dictionary_copy.get_id(token)
			)

		# Test that the vocabulary size is as expected
		self.assertEqual(
			len(unigram_dictionary_copy),
			len(self.FREQUENCIES)+1
		)

		# Test that the counts for each token are correct
		for token, count in self.FREQUENCIES.items():
			token_id = unigram_dictionary.get_id(token)
			self.assertEqual(
				unigram_dictionary_copy.get_frequency(token_id),
				count
			)

		# Test that the number of tokens is as expected
		self.assertEqual(
			unigram_dictionary_copy.get_num_tokens(),
			sum(self.FREQUENCIES.values())
		)


	def test_pruning(self):


		# Make a unigram dictionary, and populate it with the corpus
		unigram_dictionary = UnigramDictionary()
		unigram_dictionary.update(self.CORPUS)

		# Ensure that the dictionary has correctly encoded the desired
		# information about the corpus.
		for token in unigram_dictionary.token_map.tokens:
			token_id = unigram_dictionary.get_id(token)
			freq = unigram_dictionary.get_frequency(token_id)
			if token == 'UNK':
				self.assertEqual(freq, 0)
			else:
				self.assertEqual(freq, self.FREQUENCIES[token])

		# Check that the dictionary knows the correct number of words.
		# Recall that this is one more than the number of unique words
		# in the corpus, because of the reserved 'UNK' word.
		num_tokens = len(self.FREQUENCIES)+1
		self.assertEqual(len(unigram_dictionary), num_tokens)

		# Prune the dictionary!
		unigram_dictionary.prune(min_frequency=6)

		# Check that two elements were dropped from the dictionary
		# ('apple' and 'pineapple').
		num_tokens -= 2
		self.assertEqual(len(unigram_dictionary), num_tokens)

		# Check that the frequences are as expected.  Counts for
		# apple and pineapple should have been attributed to UNK
		unk_freq = (
			self.FREQUENCIES['apple'] + self.FREQUENCIES['pineapple']
		)
		for token in unigram_dictionary.token_map.tokens:

			# We should not see apple or pineapple
			self.assertTrue(token not in ('apple', 'pineapple'))

			token_id = unigram_dictionary.get_id(token)
			freq = unigram_dictionary.get_frequency(token_id)
			if token == 'UNK':
				self.assertEqual(freq, unk_freq)
			else:
				self.assertEqual(freq, self.FREQUENCIES[token])



class TestTokenChooser(TestCase):
	'''
	Given a list of tokens (usually representing a sentence), and a 
	particular query token, the job of the TokenChooser is to sample
	from the nearby tokens, within a window of +/- K tokens.
	It does not yield the sampled token itself, but rather, its index.
	A complicating issue is the fact that the query token might be 
	near the beginning or end of the token list (sentence), which needs
	to be accounted for so that the sampled index doesn't produce an 
	IndexError
	'''

	def test_mid_sentence(self):
		'''
		Test token chooser's performance given that the query token 
		is far from the edges of the context
		'''

		# Seed randomness for reproducibility
		np.random.seed(1)

		# Set up some testing data.  We'll sample from this sentence
		# using the word "sufficiently" as the query word
		sentence = (
			'this is a test sentence that is sufficiently long to '
			'enable testing various edge cases'
		).split()
		query_idx = sentence.index('sufficiently')
		context_length = len(sentence)

		# Sample from +/- 5 words
		K = 5
		# Weight the probabilities like this (higher probability of 
		# sampling near the query word itself).
		counts = [1,2,3,4,5,5,4,3,2,1]

		# Make the chooser
		chooser = TokenChooser(K, counts)

		# Sample many times so we can test the sample's statistics
		num_samples = 100000
		found_counts = Counter([
			chooser.choose_token(query_idx, context_length)
			for s in range(num_samples)
		])

		# Convert counts into frequencies
		found_frequencies = dict([
			(idx, c / float(num_samples)) 
			for idx, c in found_counts.items()
		])

		# Assemble the expected frequencies.  First, the acceptable
		# outcomes are tokens that are within +/- 5.  So the acceptable
		# relative indices are:
		expected_relative_idxs = range(-5, 0) + range(1, 6)

		# And so the expected absolute indices are:
		expected_idxs = [
			rel + query_idx 
			for rel in expected_relative_idxs
		]

		# Calculate the expected frequency that each index should appear
		# in the sample
		total = float(sum(counts))
		expected_frequencies = dict(
			[(idx, c / total) for idx, c in zip(expected_idxs, counts)]
		)

		# First, make sure that the sample indices are the ones expected
		self.assertEqual(
			set(found_frequencies.keys()),
			set(expected_frequencies.keys())
		)

		# Then check that the indices in the sample arise with 
		# approximately the frequencies expected
		tolerance = 0.002
		for idx in expected_frequencies:
			diff = abs(expected_frequencies[idx] - found_frequencies[idx])
			self.assertTrue(diff < tolerance)


	def test_near_beginning(self):
		'''
		Test token chooser's performance given that the query token 
		is close to the beginning of the context
		'''

		# Seed randomness for reproducibility
		np.random.seed(1)

		# Set up some testing data.  We'll sample from this sentence
		# using the word "a" as the query word
		sentence = (
			'this is a test sentence that is sufficiently long to '
			'enable testing various edge cases'
		).split()
		query_idx = sentence.index('a')
		context_length = len(sentence)

		# Sample from +/- 5 words
		K = 5
		# Weight the probabilities like this (higher probability of 
		# sampling near the query word itself).
		counts = [1,2,3,4,5,5,4,3,2,1]

		# Make the chooser
		chooser = TokenChooser(K, counts)

		# Sample many times so we can test the sample's statistics
		num_samples = 100000
		found_counts = Counter([
			chooser.choose_token(query_idx, context_length)
			for s in range(num_samples)
		])

		# Convert counts into frequencies
		found_frequencies = dict([
			(idx, c / float(num_samples)) 
			for idx, c in found_counts.items()
		])

		# Assemble the expected frequencies.  First, the acceptable
		# outcomes are tokens that are within +/- 5.  But because the
		# query token is near the start of the context, there are actually
		# only two tokens available before it
		expected_relative_idxs = [-2, -1] + range(1, 6)

		# And so the expected absolute indices are:
		expected_idxs = [
			rel + query_idx 
			for rel in expected_relative_idxs
		]

		# Calculate the expected frequency that each index should appear
		# in the sample
		relative_frequencies = [4,5,5,4,3,2,1]
		total = float(sum(relative_frequencies))
		expected_frequencies = dict([
			(idx, c / total) 
			for idx, c in zip(expected_idxs, relative_frequencies)
		])

		# First, make sure that the sample indices are the ones expected
		self.assertEqual(
			set(found_frequencies.keys()),
			set(expected_frequencies.keys())
		)

		# Then check that the indices in the sample arise with 
		# approximately the frequencies expected
		tolerance = 0.003
		for idx in expected_frequencies:
			diff = abs(expected_frequencies[idx] - found_frequencies[idx])
			self.assertTrue(diff < tolerance)


	def test_near_end(self):
		'''
		Test token chooser's performance given that the query token 
		is close to the end of the context
		'''

		# Seed randomness for reproducibility
		np.random.seed(1)

		# Set up some testing data.  We'll sample from this sentence
		# using the word "cases" as the query word
		sentence = (
			'this is a test sentence that is sufficiently long to '
			'enable testing various edge cases'
		).split()
		query_idx = sentence.index('cases')
		context_length = len(sentence)

		# Sample from +/- 5 words
		K = 5
		# Weight the probabilities like this (higher probability of 
		# sampling near the query word itself).
		counts = [1,2,3,4,5,5,4,3,2,1]

		# Make the chooser
		chooser = TokenChooser(K, counts)

		# Sample many times so we can test the sample's statistics
		num_samples = 100000
		found_counts = Counter([
			chooser.choose_token(query_idx, context_length)
			for s in range(num_samples)
		])

		# Convert counts into frequencies
		found_frequencies = dict([
			(idx, c / float(num_samples)) 
			for idx, c in found_counts.items()
		])

		# Assemble the expected frequencies.  First, the acceptable
		# outcomes are tokens that are within +/- 5.  But because the
		# query token is near the start of the context, there are actually
		# only two tokens available before it
		expected_relative_idxs = range(-5, 0)

		# And so the expected absolute indices are:
		expected_idxs = [
			rel + query_idx 
			for rel in expected_relative_idxs
		]

		# Calculate the expected frequency that each index should appear
		# in the sample
		relative_frequencies = [1,2,3,4,5]
		total = float(sum(relative_frequencies))
		expected_frequencies = dict([
			(idx, c / total) 
			for idx, c in zip(expected_idxs, relative_frequencies)
		])

		# First, make sure that the sample indices are the ones expected
		self.assertEqual(
			set(found_frequencies.keys()),
			set(expected_frequencies.keys())
		)

		# Then check that the indices in the sample arise with 
		# approximately the frequencies expected
		tolerance = 0.003
		for idx in expected_frequencies:
			diff = abs(expected_frequencies[idx] - found_frequencies[idx])
			self.assertTrue(diff < tolerance)


	def test_short_context(self):
		'''
		Test token chooser's performance given that context is short
		'''

		# Seed randomness for reproducibility
		np.random.seed(1)

		# Set up some testing data.  We'll sample from this sentence
		# using the word "This" as the query word
		sentence = 'This is short'.split()
		query_idx = sentence.index('This')
		context_length = len(sentence)

		# Sample from +/- 5 words
		K = 5
		# Weight the probabilities like this (higher probability of 
		# sampling near the query word itself).
		counts = [1,2,3,4,5,5,4,3,2,1]

		# Make the chooser
		chooser = TokenChooser(K, counts)

		# Sample many times so we can test the sample's statistics
		num_samples = 100000
		found_counts = Counter([
			chooser.choose_token(query_idx, context_length)
			for s in range(num_samples)
		])

		# Convert counts into frequencies
		found_frequencies = dict([
			(idx, c / float(num_samples)) 
			for idx, c in found_counts.items()
		])

		# Assemble the expected frequencies.  First, the acceptable
		# outcomes are tokens that are within +/- 5.  But because the
		# query token is near the start of the context, there are actually
		# only two tokens available before it
		expected_relative_idxs = [1,2]

		# And so the expected absolute indices are:
		expected_idxs = [
			rel + query_idx 
			for rel in expected_relative_idxs
		]

		# Calculate the expected frequency that each index should appear
		# in the sample
		relative_frequencies = [5,4]
		total = float(sum(relative_frequencies))
		expected_frequencies = dict([
			(idx, c / total) 
			for idx, c in zip(expected_idxs, relative_frequencies)
		])

		# First, make sure that the sample indices are the ones expected
		self.assertEqual(
			set(found_frequencies.keys()),
			set(expected_frequencies.keys())
		)

		# Then check that the indices in the sample arise with 
		# approximately the frequencies expected
		tolerance = 0.003
		for idx in expected_frequencies:
			diff = abs(expected_frequencies[idx] - found_frequencies[idx])
			self.assertTrue(diff < tolerance)



class TestCounterSampler(TestCase):

	def test_sampling(self):
		'''
		Test basic function of assigning counts, and then sampling from
		The distribution implied by those counts.
		'''
		counts = range(1,6)
		counter_sampler = CounterSampler()
		counter_sampler.update(counts)

		# Test asking for a single sample (where no shape tuple supplied)
		single_sample = counter_sampler.sample()
		self.assertTrue(type(single_sample) is np.int64)

		# Test asking for an array of samples (by passing a shape tuple)
		shape = (2,3,5)
		array_sample = counter_sampler.sample(shape)
		self.assertTrue(type(array_sample) is np.ndarray)
		self.assertTrue(array_sample.shape == shape)


	def test_add_function(self):
		'''
		Make sure that the add function is working correctly.  
		CounterSampler stores counts as list, wherein the value at 
		position i of the list encodes the number of counts seen for 
		outcome i.

		Counts are added by passing the outcome's index into 
		CounterSampler.add()
		which leads to position i of the counts list to be incremented.
		If position i doesn't exist, it is created.  If the counts list
		had only j elements before, and a count is added for position
		i, with i much greater than j, then many elements are created 
		between i and j, and are provisionally initialized with zero 
		counts.

		Ensure that is done properly
		'''

		counter_sampler = CounterSampler()
		self.assertEqual(counter_sampler.counts, [])

		outcome_to_add = 6
		counter_sampler.add(outcome_to_add)
		expected_counts = [0]*(outcome_to_add) + [1]
		self.assertEqual(counter_sampler.counts, expected_counts)

		# Now ensure the underlying sampler can tolerate a counts list
		# containing zeros, and that the sampling statistics is as 
		# expected.  We expect that the only outcome that should turn up 
		# is outcome 6, since it has all the probability mass.  Check that.
		counter = Counter(counter_sampler.sample((100000,))) # should be 
															 # all 6's
		total = float(sum(counter.values()))
		found_normalized = [
			counter[i] / total for i in range(outcome_to_add+1)
		]

		# Make an list of the expected fractions by which each outcome
		# should be observed, in the limit of infinite sample
		expected_normalized = expected_counts

		# Check if each outcome was observed with a fraction that is within
		# 0.005 of the expected fraction
		self.assertEqual(found_normalized, expected_normalized)


	def test_counter_sampler_statistics(self):
		'''
		This tests that the sampler really does produce results whose
		statistics match those requested by the counts vector
		'''
		# Seed numpy's random function to make the test reproducible
		np.random.seed(1)

		# Make a sampler with probabilities proportional to counts
		counts = range(1,6)
		counter_sampler = CounterSampler()
		for outcome, count in enumerate(counts):
			counter_sampler.update([outcome]*count)

		# Draw one hundred thousand samples, then total up the fraction of
		# each outcome obseved
		counter = Counter(counter_sampler.sample((100000,)))
		total = float(sum(counter.values()))
		found_normalized = [
			counter[i] / total for i in range(len(counts))
		]

		# Make an list of the expected fractions by which each outcome
		# should be observed, in the limit of infinite sample
		total_in_expected = float(sum(counts))
		expected_normalized = [
			c / total_in_expected for c in counts
		]

		# Check if each outcome was observed with a fraction that is within
		# 0.005 of the expected fraction
		close = [
			abs(f - e) < 0.005 
			for f,e in zip(found_normalized, expected_normalized)
		]
		self.assertTrue(all(close))


	def test_save_load(self):

		fname = 'test-data/test-counter-sampler/test-counter-sampler.gz'

		# Make a sampler with probabilities proportional to counts
		counts = range(1,6)
		counter_sampler = CounterSampler()
		for outcome, count in enumerate(counts):
			counter_sampler.update([outcome]*count)

		counter_sampler.save(fname)

		new_counter_sampler = CounterSampler()
		new_counter_sampler.load(fname)
		self.assertEqual(new_counter_sampler.counts, counts)


class TestMultinomialSampler(TestCase):

	def test_multinomial_sampler(self):
		counts = range(1,6)
		sampler = MultinomialSampler(counts)

		# Test asking for a single sample (where no shape tuple supplied)
		single_sample = sampler.sample()
		self.assertTrue(type(single_sample) is np.int64)

		# Test asking for an array of samples (by passing a shape tuple)
		shape = (2,3,5)
		array_sample = sampler.sample(shape)
		self.assertTrue(type(array_sample) is np.ndarray)
		self.assertTrue(array_sample.shape == shape)


	def test_multinomial_sampler_stats(self):
		'''
		This tests that the sampler really does produce results whose
		statistics match those requested by the counts vector
		'''
		# Seed numpy's random function to make the test reproducible
		np.random.seed(1)

		# Make a sampler with probabilities proportional to counts
		counts = range(1,6)
		sampler = MultinomialSampler(counts)

		# Draw one hundred thousand samples, then total up the fraction of
		# each outcome obseved
		counter = Counter(sampler.sample((100000,)))
		total = float(sum(counter.values()))
		found_normalized = [
			counter[i] / total for i in range(len(counts))
		]

		# Make an list of the expected fractions by which each outcome
		# should be observed, in the limit of infinite sample
		total_in_expected = float(sum(counts))
		expected_normalized = [
			c / total_in_expected for c in counts
		]

		# Check if each outcome was observed with a fraction that is within
		# 0.005 of the expected fraction
		close = [
			abs(f - e) < 0.005 
			for f,e in zip(found_normalized, expected_normalized)
		]
		self.assertTrue(all(close))



class TestTokenMap(TestCase):

	TOKENS = ['apple', 'pear', 'banana', 'orange']

	def test_token_map(self):

		token_map = TokenMap(on_unk=SILENT)

		for idx, fruit in enumerate(self.TOKENS):
			# Ensure that ids are assigned in an auto-incrementing way
			# starting from 1 (0 is reserved for the UNK token)
			self.assertEqual(token_map.add(fruit), idx+1)

		for idx, fruit in enumerate(self.TOKENS):
			# Ensure that idxs are stable and retrievable with 
			# TokenMap.get_id()
			self.assertEqual(token_map.get_id(fruit), idx+1)

			# Ensure that we can look up the token using the id
			self.assertEqual(token_map.get_token(idx+1), fruit)

		# Ensure the token_map knows its own length
		self.assertEqual(len(token_map), len(self.TOKENS)+1)

		# Asking for ids of non-existent tokens returns the UNK token_id
		self.assertEqual(token_map.get_id('no-exist'), 0)

		# Asking for the token at 0 returns 'UNK'
		self.assertEqual(token_map.get_token(0), 'UNK')

		# Asking for token at non-existent idx raises IndexError
		with self.assertRaises(IndexError):
			token_map.get_token(99)


	def test_raise_error_on_unk(self):
		'''
		If the token_map is constructed passing 
			on_unk=TokenMap.ERROR
		then calling get_id() or get_ids() will throw a KeyError if one
		of the supplied tokens isn't in the token_map.  (Normally it 
		would return 0, which is a token id reserved for 'UNK' -- any
		unknown token).
		'''

		token_map = TokenMap(on_unk=ERROR)
		token_map.update(self.TOKENS)

		with self.assertRaises(KeyError):
			token_map.get_id('no-exist')

		with self.assertRaises(KeyError):
			token_map.get_ids(['apple', 'no-exist'])



	def test_token_map_plural_functions(self):

		token_map = TokenMap(on_unk=SILENT)

		# In these assertions, we offset the expected list of ids by
		# 1 because the 0th id in token_map is reserved for the UNK
		# token

		# Ensure that update works
		ids = token_map.update(self.TOKENS)
		self.assertEqual(ids, range(1, len(self.TOKENS)+1))

		# Ensure that get_ids works
		self.assertEqual(
			token_map.get_ids(self.TOKENS),
			range(1, len(self.TOKENS)+1)
		)

		# Ensure that get_tokens works
		self.assertEqual(
			token_map.get_tokens(range(1, len(self.TOKENS)+1)),
			self.TOKENS
		)

		# Asking for ids of non-existent tokens raises KeyError
		self.assertEqual(
			token_map.get_ids(['apple', 'no-exist']),
			[self.TOKENS.index('apple')+1, 0]
		)

		# Asking for token at 0 returns the 'UNK' token
		self.assertEqual(
			token_map.get_tokens([3,0]),
			[self.TOKENS[3-1], 'UNK']
		)

		# Asking for token at non-existent idx raises IndexError
		with self.assertRaises(IndexError):
			token_map.get_tokens([1,99])


	def test_save_load(self):
		token_map = TokenMap(on_unk=SILENT)
		token_map.update(self.TOKENS)
		token_map.save('test-data/test-token-map/test-token-map.gz')

		token_map_copy = TokenMap(on_unk=SILENT)
		token_map_copy.load(
			'test-data/test-token-map/test-token-map.gz'
		)
		self.assertEqual(
			token_map_copy.get_ids(self.TOKENS),
			range(1, len(self.TOKENS)+1)
		)
		self.assertEqual(len(token_map_copy), len(self.TOKENS)+1)



class TestNoiseContrast(TestCase):
	def test_noise_contrast(self):
		signal_input = T.dvector()
		noise_input = T.dmatrix()
		loss = noise_contrast(signal_input, noise_input)

		f = function([signal_input, noise_input], loss)

		test_signal = np.array([.6, .7, .8])
		test_noise = np.array([[.2, .3, .1],[.5, .6, .7]])
		test_loss = f(test_signal, test_noise)

		expected_objective = (
			np.log(test_signal).sum() + np.log(1-test_noise).sum()
		)
		expected_loss = -expected_objective / float(len(test_signal))

		#print 'expected_loss:'
		#print expected_loss
		#print
		#print 'test_loss:'
		#print test_loss
		#print

		self.assertAlmostEqual(test_loss, expected_loss)



class TestMinibatcher(TestCase):

	def setUp(self):

		# Define some parameters to be used in construction 
		# Minibatcher
		self.files = [
			'test-data/test-corpus/003.tsv',
			'test-data/test-corpus/004.tsv'
		]
		self.batch_size = 5
		self.noise_ratio = 15
		self.t = 0.03

		# Make a minibatch generator
		self.generator = Word2VecMinibatcher(
			files=self.files,
			t=self.t,
			batch_size=self.batch_size,
			noise_ratio=self.noise_ratio,
			verbose=False
		)

		# Make another Word2VecMinibatcher, and pre-load this one with 
		# token_map and counter_sampler distribution information.
		self.preloaded_generator = Word2VecMinibatcher(
			files=self.files,
			t=self.t,
			batch_size=self.batch_size,
			noise_ratio=self.noise_ratio,
			verbose=False
		)
		self.preloaded_generator.load('test-data/minibatch-generator-test')


	def test_prepare(self):
		'''
		Check that Word2VecMinibatcher.prepare() properly makes a 
		UnigramDictionary that reflects the corpus.
		'''
		self.generator.prepare()
		d = self.generator.unigram_dictionary

		# Make sure that all of the expected tokens are found in the 
		# unigram_dictionary, and that their frequency in the is correct.
		tokens = []
		for filename in self.files:
			for add_tokens in default_parse(filename):
				tokens.extend(add_tokens)

		counts = Counter(tokens)
		for token in tokens:
			token_id = d.get_id(token)
			count = d.get_frequency(token_id)
			self.assertEqual(count, counts[token])


	def test_minibatches(self):
		'''
		Make sure that the minibatches are the correct size, that 
		signal query- and contexts-words are always within 5 tokens of
		one another and come from the same sentence.
		'''
		# Ensure reproducibility in this stochastic test
		np.random.seed(1)

		# Before looking at the minibatches, we need to determine what 
		# query-context pairs are possible.
		# To do that, first read in the corpus, and break it into lines
		# and tokens
		lines = []

		# Go through the corpus and get all the token ids as a list
		tokenized_lines = []
		for filename in self.files:
			for tokens in default_parse(filename):
				tokenized_lines.append(tokens)

		# Now iterate through the lines, noting what tokens arise within
		# one another's contexts.  Build a lookup table providing the set
		# of token_ids that arose in the context of each given token_id
		legal_pairs = defaultdict(set)
		d = self.preloaded_generator.unigram_dictionary
		for line in tokenized_lines:
			token_ids = d.get_ids(line)
			for i, token_id in enumerate(token_ids):
				low = max(0, i-5)
				legal_context = token_ids[low:i] + token_ids[i+1:i+6]
				legal_pairs[token_id].update(legal_context)

		# finally, add UNK to the legal pairs
		legal_pairs[0] = set([0])

		for signal_batch, noise_batch in self.preloaded_generator:

			self.assertEqual(len(signal_batch), self.batch_size)
			self.assertEqual(
				len(noise_batch), self.batch_size * self.noise_ratio
			)

			# Ensure that all of the signal examples are actually valid
			# samples from the corpus
			for query_token_id, context_token_id in signal_batch:
				self.assertTrue(
					context_token_id in legal_pairs[query_token_id]
				)

	def test_token_discarding(self):

		# Ensure reproducibility for the test
		np.random.seed(1)

		# Get the preloaded generator and its unigram_dictionary
		self.preloaded_generator
		d = self.preloaded_generator.unigram_dictionary

		# Go through the corpus and get all the token ids as a list
		token_ids = []
		for filename in self.files:
			for tokens in default_parse(filename):
				token_ids.extend(d.get_ids(tokens))

		# Run through the tokens, evaluating 
		# Word2VecMinibatcher.do_discard() on each.  Keep track of all 
		# "discarded" tokens for which do_discard() returns True
		discarded = []
		num_replicates = 100
		for replicates in range(num_replicates):
			for token_id in token_ids:
				if self.preloaded_generator.do_discard(token_id):
					discarded.append(token_id)

		# Count the tokens, and the discarded tokens.  
		discarded_counts = Counter(discarded)
		token_counts = Counter(token_ids)

		# Compute the frequency of the word "the", and the frequency
		# with which it was discarded
		the_id = d.get_id('the')
		num_the_appears = token_counts[the_id]
		the_frequency = num_the_appears/float(len(token_ids))
		num_the_discarded = discarded_counts[the_id]
		frequency_of_discard = (
			num_the_discarded / float(num_the_appears * num_replicates)
		)

		# What was actually the most discarded token?  It should be "the"
		most_discarded_id, num_most_discarded = (
			discarded_counts.most_common()[0]
		)
		self.assertEqual(most_discarded_id, the_id)

		# What was the expected frequency with which "the" would be 
		# discarded?  Assert it is close to the actual discard rate.
		expected_frequency = 1 - np.sqrt(self.t / the_frequency)
		tolerance = 0.015
		self.assertTrue(
			abs(expected_frequency - frequency_of_discard) < tolerance
		)



class TestWord2VecOnCorpus(TestCase):
	'''
	This tests the Word2Vec end-to-end functionality applied to a text 
	corpus.
	'''

	def test_word2vec_on_corpus(self):

		# Seed randomness to make the test reproducible
		np.random.seed(1)

		word2vec_embedder, dictionary = word2vec(
			files=['test-data/test-corpus/numbers-long.txt'],
			num_epochs=1,
			t=1, 
			batch_size = 10,
			num_embedding_dimensions=5,
			verbose=False,
		)

		W, C = word2vec_embedder.get_param_values()
		dots = usigma(np.dot(W,C.T))

		# Based on the construction of the corpus, the following 
		# context embeddings should match the query at right and be 
		# the highest value in the product of the embedding matrices
		# Note that token 0 is reserved for UNK.  It's embedding stays
		# near the randomly initialized value, tending to yield of 0.5
		# which is high overall, so it turns up as a "good match" to any
		# word
		expected_tops = [
			[0,2,3], # these contexts are good match to query 1
			[0,1,3], # these contexts are good match to query 2 
			[0,1,2], # these contexts are good match to query 3 
			[0,5,6], # these contexts are good match to query 4 
			[0,4,6], # these contexts are good match to query 5 
			[0,4,5], # these contexts are good match to query 6 
			[0,8,9], # these contexts are good match to query 7 
			[0,7,9], # these contexts are good match to query 8 
			[0,7,8], # these contexts are good match to query 9 
			[0,11,12], # these contexts are good match to query 10
			[0,10,12], # these contexts are good match to query 11 
			[0,10,11]  # these contexts are good match to query 12 
		]

		for i in range(1, 3*4+1):
			top3 = sorted(
				enumerate(dots[i]), key=lambda x: x[1], reverse=True
			)[:3]
			top3_positions = [t[0] for t in top3]
			self.assertItemsEqual(top3_positions, expected_tops[i-1])

		print repr(word2vec_embedder.embed([1,2,3,4]))




class TestWord2Vec(TestCase):

	'''
	This tests comnponent Word2Vec functionality by supplying 
	synthetic numerical data into its components, to make sure that 
	the solutions are mathematically correct.  It doesn't test iteration
	over an actual text corpus, which is tested by TestWord2VecOnCorpus.
	'''

	def setUp(self):

		self.VOCAB_SIZE = 3
		self.NUM_EMBEDDING_DIMENSIONS = 4

		# Make a query input vector.  This holds indices that represent
		# words in the vocabulary.  For the test we have just three words
		# in the vocabulary
		self.TEST_INPUT = np.array(
			[[i / self.VOCAB_SIZE, i % self.VOCAB_SIZE] for i in range(9)]
		).astype('int32')

		# Artificially adopt this word embedding matrix for query words
		self.QUERY_EMBEDDING = np.array([
			[-0.04576914, -0.26519672, -0.06857708, -0.23748968],
			[ 0.08540803,  0.32099229, -0.19136694, -0.48263541],  
			[-0.33319689,  0.26062664,  0.06826347, -0.39083191]
		])

		# Artificially adopt this word embedding matrix for context words
		self.CONTEXT_EMBEDDING = np.array([
			[-0.29474795, -0.2559814 , -0.04503929,  0.35159791],
			[ 0.00963128,  0.22368461,  0.44933862,  0.48584304], 
			[ 0.05338832, -0.22895403, -0.08288041, -0.47226618],
		])

		# This is the expected result if we use the TEST_QUERY_INPUT,
		# and TEST_CONTEXT_INPUT as inputs and if we use, as 
		# initializations for the embedding weights, 
		#the QUERY_EMBEDDING and CONTEXT_EMBEDDING
		self.EXPECTED_DOTS = np.array([
			[ 0.1769407 , -0.2194758 ,  0.04109232,  0.1769407 , 
				-0.2194758 , 0.04109232],
			[-0.09276679,  0.21486867, -0.25680422, -0.09276679,  
				0.21486867, -0.25680422]
		])
 
		# This is the expected result if we use the TEST_QUERY_INPUT
		# and TEST_CONTEXT_INPUT after modifying word_embedding matrix
		# by adding 10 elementwise
		self.ALT_EXPECTED_DOTS = np.array([
			[2.4208559 ,  11.1055778 , -10.86122348,   2.4208559 ,
				11.1055778 , -10.86122348],
			[11.23228681, -10.68744713,   1.98711098,  11.23228681,
				-10.68744713,   1.98711098]
		])


	# TODO: finish this
	def test_save_load(self):

		input_var = T.imatrix('input_var')

		embedder = Word2VecEmbedder(
			input_var,
			batch_size=len(self.TEST_INPUT),
			vocabulary_size=self.VOCAB_SIZE,
			num_embedding_dimensions=self.NUM_EMBEDDING_DIMENSIONS,
			word_embedding_init=self.QUERY_EMBEDDING,
			context_embedding_init=self.CONTEXT_EMBEDDING
		)

		W, C = embedder.get_param_values()


	def test_noise_contrastive_learning(self):

		# Seed randomness to make the test reproducible
		np.random.seed(1)

		# Predifine the size of batches and the embedding
		batch_size = 160
		vocab_size = 10
		num_embedding_dimensions = 5

		# Define the input theano variables
		signal_input = T.imatrix('query_input')
		noise_input = T.imatrix('noise_input')


		# Make a NoiseContraster, and get the combined input
		noise_contraster = NoiseContraster(signal_input, noise_input)
		combined_input = noise_contraster.get_combined_input()

		# Make a Word2VecEmbedder object, feed it the combined input
		word2vec_embedder = Word2VecEmbedder(
			combined_input,
			batch_size,
			vocab_size,
			num_embedding_dimensions,
		)

		# Get the params and output from the word2vec embedder, feed that
		# back to the noise_contraster to get the training function
		combined_output = word2vec_embedder.get_output()
		params = word2vec_embedder.get_params()
		train = noise_contraster.get_train_func(combined_output, params)


		# Make the positive input.  First component of each example is
		# the query input, and second component is the context.  In the 
		# final embeddings that are learned, dotting these rows and columns
		# respectively from the query and context embedding matrices should
		# give higher values than any other row-column dot products.
		test_positive_input = np.array([
			[0,2],
			[1,3],
			[2,0],
			[3,1],
			[4,6],
			[5,7],
			[6,4],
			[7,5],
			[8,9],
			[9,8]
		]).astype('int32')
		
		num_replicates = 5
		num_epochs = 3000
		embedding_products = []
		W, C = word2vec_embedder.get_params()
		start = time.time()
		for rep in range(num_replicates):
			W.set_value(np.random.normal(
				0, 0.01, (vocab_size, num_embedding_dimensions)
			).astype(dtype='float32'))
			C.set_value(np.random.normal(
				0, 0.01, (vocab_size, num_embedding_dimensions)
			).astype('float32'))
			#print '\t***'
			for epoch in range(num_epochs):

				# Sample new noise examples every epoch (this is better 
				# than fixing the noise once at the start).
				# Provide 15 negative examples for each query word
				test_negative_input = np.array([
					[i / 10, np.random.randint(0,10)] for i in range(100)
				]).astype('int32')

				loss = train(
					test_positive_input, test_negative_input
				)
				#print loss

			embedding_product = np.dot(W.get_value(), C.get_value().T)
			embedding_products.append(usigma(embedding_product))

		mean_embedding_products = np.mean(embedding_products, axis=0)
		#print np.round(mean_embedding_products, 2)

		# We expect that the embeddings will allocate the most probability
		# to the contexts that were provided for words in the toy data.
		# We always provided a single context via batch_contexts 
		# (e.g. context 2 provided for word 0), so we expect these contexts
		# to be the maximum.
		expected_max_prob_contexts = test_positive_input[:,1]
		self.assertTrue(np.array_equal(
			np.argmax(mean_embedding_products, axis=1),
			expected_max_prob_contexts
		))

		# The dot product of a given word embedding and context embedding
		# have an interpretation as the probability that that word and
		# context derived from the toy data instead of the noise.
		# See equation 3 in Noise-Contrastive Estimation of Unnormalized 
		# Statistical Models, with Applications to Natural Image 
		# StatisticsJournal of Machine Learning Research 13 (2012), 
		# pp.307-361.
		# That shows the probability should be around 0.5
		# Since the actual values are stocastic, we check that the 
		# average of repeated trials is within 0.25 - 0.75.
		embedding_maxima = np.max(mean_embedding_products, axis=1)
		self.assertTrue(all(
			[x > 0.25 for x in embedding_maxima]
		))
		self.assertTrue(all(
			[x < 0.75 for x in embedding_maxima]
		))
		#print 'average weight to correct pairs:', np.mean(
		#	embedding_maxima
		#)
		#print 'elapsed time:', time.time() - start



	# TODO: re-implement this test without the defunct Word2Vec class
	#def test_learning(self):

	#	positive_input = T.imatrix('query_input')
	#	negative_input = T.imatrix('noise_input')

	#	# Predifine the size of batches and the embedding
	#	batch_size = 160
	#	vocab_size = 10
	#	num_embedding_dimensions = 5

	#	# Make a Word2Vec object
	#	word2vec = Word2Vec(
	#		batch_size,
	#		vocabulary_size=vocab_size,
	#		num_embedding_dimensions=num_embedding_dimensions,
	#		learning_rate=0.1,
	#		verbose=False
	#	)

	#	# Make the positive input.  First component of each example is
	#	# the query input, and second component is the context.  In the 
	#	# final embeddings that are learned, dotting these rows and columns
	#	# respectively from the query and context embedding matrices should
	#	# give higher values than any other row-column dot products.
	#	test_positive_input = np.array([
	#		[0,2],
	#		[1,3],
	#		[2,0],
	#		[3,1],
	#		[4,6],
	#		[5,7],
	#		[6,4],
	#		[7,5],
	#		[8,9],
	#		[9,8]
	#	]).astype('int32')
	#	
	#	num_replicates = 5
	#	num_epochs = 3000
	#	embedding_products = []
	#	W, C = word2vec.get_params()
	#	start = time.time()
	#	for rep in range(num_replicates):
	#		W.set_value(np.random.normal(
	#			0, 0.01, (vocab_size, num_embedding_dimensions)
	#		).astype(dtype='float32'))
	#		C.set_value(np.random.normal(
	#			0, 0.01, (vocab_size, num_embedding_dimensions)
	#		).astype('float32'))
	#		#print '\t***'
	#		for epoch in range(num_epochs):

	#			# Sample new noise examples every epoch (this is better than
	#			# fixing the noise once at the start).
	#			# Provide 15 negative examples for each query word
	#			test_negative_input = np.array([
	#				[i / 10, random.randint(0,9)] for i in range(100)
	#			]).astype('int32')

	#			loss = word2vec.train(
	#				test_positive_input, test_negative_input
	#			)
	#			#print loss

	#		embedding_product = np.dot(W.get_value(), C.get_value().T)
	#		embedding_products.append(usigma(embedding_product))

	#	mean_embedding_products = np.mean(embedding_products, axis=0)
	#	#print np.round(mean_embedding_products, 2)

	#	# We expect that the embeddings will allocate the most probability
	#	# to the contexts that were provided for words in the toy data.
	#	# We always provided a single context via batch_contexts 
	#	# (e.g. context 2 provided for word 0), so we expect these contexts
	#	# to be the maximum.
	#	expected_max_prob_contexts = test_positive_input[:,1]
	#	self.assertTrue(np.array_equal(
	#		np.argmax(mean_embedding_products, axis=1),
	#		expected_max_prob_contexts
	#	))

	#	# The dot product of a given word embedding and context embedding
	#	# have an interpretation as the probability that that word and
	#	# context derived from the toy data instead of the noise.
	#	# See equation 3 in Noise-Contrastive Estimation of Unnormalized 
	#	# Statistical Models, with Applications to Natural Image 
	#	# StatisticsJournal of Machine Learning Research 13 (2012), 
	#	# pp.307-361.
	#	# That shows the probability should be around 0.5
	#	# Since the actual values are stocastic, we check that the 
	#	# average of repeated trials is within 0.25 - 0.75.
	#	embedding_maxima = np.max(mean_embedding_products, axis=1)
	#	self.assertTrue(all(
	#		[x > 0.25 for x in embedding_maxima]
	#	))
	#	self.assertTrue(all(
	#		[x < 0.75 for x in embedding_maxima]
	#	))
	#	#print 'average weight to correct pairs:', np.mean(
	#	#	embedding_maxima
	#	#)
	#	#print 'elapsed time:', time.time() - start


	# TODO: re-implement this test without the defunct Word2Vec class
	#def test_positive_negative_separation(self):

	#	vocabulary_size=100000,
	#	num_embedding_dimensions=500,
	#	word_embedding_init=Normal(), 
	#	context_embedding_init=Normal(),

	#	embedder = Word2Vec(
	#		batch_size=len(self.TEST_INPUT),
	#		vocabulary_size=self.VOCAB_SIZE,
	#		num_embedding_dimensions=self.NUM_EMBEDDING_DIMENSIONS,
	#		word_embedding_init=self.QUERY_EMBEDDING,
	#		context_embedding_init=self.CONTEXT_EMBEDDING
	#	)

	#	# Get Word2Vec's internal symbolic theano variables
	#	positive_input = embedder.positive_input
	#	negative_input = embedder.negative_input
	#	positive_output = embedder.positive_output
	#	negative_output = embedder.negative_output

	#	# Compile the function connecting the symbolic inputs and outputs
	#	# So that we can test the separation of the signal and noise 
	#	# channels after they are processed by the underlying 
	#	# Word2VecEmbedder
	#	f = function(
	#		[positive_input, negative_input],
	#		[positive_output, negative_output]
	#	)

	#	# Calculate the positive and negative output
	#	test_positive_input = self.TEST_INPUT[:6]
	#	test_negative_input = self.TEST_INPUT[6:]
	#	test_positive_output, test_negative_output = f(
	#		test_positive_input, test_negative_input
	#	)

	#	# Calculate the expected embeddings and output
	#	expected_query_embeddings = np.repeat(
	#		self.QUERY_EMBEDDING, 3, axis=0
	#	)
	#	expected_context_embeddings = np.tile(
	#		self.CONTEXT_EMBEDDING, (3,1)
	#	)
	#	expected_output = usigma(np.dot(
	#		expected_query_embeddings, expected_context_embeddings.T
	#	)).diagonal()
	#	expected_positive_output = expected_output[:6]
	#	expected_negative_output = expected_output[6:]

	#	#print 'test_positive_output:'
	#	#print test_positive_output
	#	#print
	#	#print 'expected_positive_output:'
	#	#print expected_positive_output
	#	#print
	#	#print 'test_negative_output:'
	#	#print test_negative_output
	#	#print 
	#	#print 'expected_negative_output:'
	#	#print expected_negative_output
	#	#print 

	#	# Check for equality between all found and expected values
	#	self.assertTrue(np.allclose(
	#		test_positive_output, expected_positive_output
	#	))
	#	self.assertTrue(np.allclose(
	#		test_negative_output, expected_negative_output
	#	))


	def test_Word2VecEmbedder(self):

		input_var = T.imatrix('input_var')

		embedder = Word2VecEmbedder(
			input_var,
			batch_size=len(self.TEST_INPUT),
			vocabulary_size=self.VOCAB_SIZE,
			num_embedding_dimensions=self.NUM_EMBEDDING_DIMENSIONS,
			word_embedding_init=self.QUERY_EMBEDDING,
			context_embedding_init=self.CONTEXT_EMBEDDING
		)

		query_embedding = embedder.query_embedding
		context_embedding = embedder.context_embedding
		dots = embedder.get_output()

		f = function([input_var], query_embedding)
		g = function([input_var], context_embedding)
		h = function([input_var], dots)
		
		# Calculate the embeddings and the output
		query_embeddings = f(self.TEST_INPUT)
		context_embeddings = g(self.TEST_INPUT)
		test_output = h(self.TEST_INPUT)

		# Calculate the expected embeddings and output
		expected_query_embeddings = np.repeat(
			self.QUERY_EMBEDDING, 3, axis=0
		)
		expected_context_embeddings = np.tile(
			self.CONTEXT_EMBEDDING, (3,1)
		)
		expected_output = usigma(np.dot(
			expected_query_embeddings, expected_context_embeddings.T
		)).diagonal()

		#print 'found query embeddings:'
		#print query_embeddings
		#print
		#print 'expected query embeddings:'
		#print expected_query_embeddings
		#print
		#print 'found context embeddings:'
		#print context_embeddings
		#print 
		#print 'expected context embeddings:'
		#print expected_context_embeddings
		#print 
		#print 'found output:'
		#print test_output
		#print
		#print 'expected output:'
		#print expected_output

		# Check for equality between all found and expected values
		self.assertTrue(np.allclose(
			query_embeddings, expected_query_embeddings
		))
		self.assertTrue(np.allclose(
			context_embeddings, expected_context_embeddings
		))
		self.assertTrue(np.allclose(test_output, expected_output))



if __name__=='__main__': 
	main()
	



import theano.tensor as T
from theano import function
import lasagne


def noise_contrast(signal, noise, scale=True):
	'''
	Takes the theano symbolic variables `signal` and `noise`, whose
	elements are interpreted as probabilities, and creates a loss
	function which rewards large probabilities in signal and penalizes 
	large probabilities in noise.
	
	`signal` and `noise` can have any shape.  Their contributions will be 
	summed over all dimensions.

	If `scale` is true, then scale the loss function by the size of the
	signal tensor --- i.e. divide by the signal batch size.  This makes
	the scale of the loss function invariant to changes in batch size
	'''

	signal_score = T.log(signal).sum()
	noise_score = T.log(1-noise).sum()
	objective = signal_score + noise_score
	loss = -objective

	loss = loss / signal.shape[0]
	
	return loss


class NoiseContraster(object):

	def __init__(
		self,
		signal_input,
		noise_input,
		learning_rate=0.1,
		momentum=0.9
	):
		self.signal_input = signal_input
		self.noise_input = noise_input
		self.learning_rate = learning_rate
		self.momentum=momentum

		self.combined = T.concatenate([
			self.signal_input, self.noise_input
		])

	def get_combined_input(self):
		return self.combined

	def get_train_func(self, output, params):

		# Split the output based on the size of the individual input streams
		self.signal_output = output[:self.signal_input.shape[0]]
		self.noise_output = output[self.signal_input.shape[0]:]

		# Construct the loss based on Noise Contrastive Estimation
		self.loss = noise_contrast(self.signal_output, self.noise_output)

		# Get the parameter updates using stochastic gradient descent with
		# nesterov momentum.  TODO: make the updates configurable
		self.updates = lasagne.updates.nesterov_momentum(
			self.loss,
			params,
			self.learning_rate,
			self.momentum
		)

		self.train = function(
			inputs=[self.signal_input, self.noise_input], 
			outputs=self.loss,
			updates=self.updates
		)

		return self.train






from noise_contrast import NoiseContraster
import os
import numpy as np
from theano import tensor as T, function
import lasagne
from minibatcher import Word2VecMinibatcher
from lasagne.layers import (
	get_output, InputLayer, EmbeddingLayer, get_all_params,
	get_all_param_values
)
from lasagne.init import Normal


def row_dot(matrixA, matrixB):
	C = (matrixA * matrixB).sum(axis=1)
	return C


def sigmoid(tensor_var):
	return 1/(1+T.exp(-tensor_var))


def word2vec(
		files=[],
		directories=[],
		skip=[],
		savedir=None,
		num_epochs=5,
		unigram_dictionary=None,
		noise_ratio=15,
		kernel=[1,2,3,4,5,5,4,3,2,1],
		t = 1.0e-5,
		batch_size = 1000,
		num_embedding_dimensions=500,
		word_embedding_init=Normal(),
		context_embedding_init=Normal(),
		learning_rate=0.1,
		momentum=0.9,
		verbose=True,
		num_example_generators=3
	):

	'''
	Helper function that handles all concerns involved in training
	A word2vec model using the approach of Mikolov et al.  It surfaces
	all of the options. 

	For customizations going beyond simply tweeking existing options and 
	hyperparameters, substitute this function by writing your own training 
	routine using the provided classes.  This function would be a starting
	point for you.
	'''

	# Make a Word2VecMinibatcher
	minibatcher = Word2VecMinibatcher(
		files=files,
		directories=directories,
		skip=skip,
		unigram_dictionary=unigram_dictionary,
		noise_ratio=noise_ratio,
		kernel=kernel,
		t=t,
		batch_size=batch_size,
		verbose=verbose,
		num_example_generators=num_example_generators
	)

	# Prpare the minibatch generator 
	# (this produces the counter_sampler stats)
	minibatcher.prepare(savedir=savedir)

	# Define the input theano variables
	signal_input = T.imatrix('query_input')
	noise_input = T.imatrix('noise_input')

	# Make a NoiseContraster, and get the combined input
	noise_contraster = NoiseContraster(signal_input, noise_input)
	combined_input = noise_contraster.get_combined_input()

	# Make a Word2VecEmbedder object, feed it the combined input
	word2vec_embedder = Word2VecEmbedder(
		input_var=combined_input,
		batch_size=batch_size,
		vocabulary_size=minibatcher.get_vocab_size(),
		num_embedding_dimensions=num_embedding_dimensions,
		word_embedding_init=word_embedding_init,
		context_embedding_init=context_embedding_init
	)

	# Get the params and output from the word2vec embedder, feed that
	# back to the noise_contraster to get the training function
	combined_output = word2vec_embedder.get_output()
	params = word2vec_embedder.get_params()
	train = noise_contraster.get_train_func(combined_output, params)

	# Iterate over the corpus, training the embeddings
	for epoch in range(num_epochs):
		if verbose:
			print 'starting epoch %d' % epoch
		losses = []
		for signal_batch, noise_batch in minibatcher:
			losses.append(train(signal_batch, noise_batch))
		if verbose:
			print '\tAverage loss: %f' % np.mean(losses)

	# Save the model (the embeddings) if savedir was provided
	if savedir is not None:
		embedings_filename = os.path.join(savedir, 'embeddings.npz')
		word2vec_embedder.save(embeddings_filename)

	# Return the trained word2vec_embedder and the dictionary mapping tokens
	# to ids
	return word2vec_embedder, minibatcher.unigram_dictionary



class Word2VecEmbedder(object):

	def __init__(
		self,
		input_var,
		batch_size,
		vocabulary_size=100000,
		num_embedding_dimensions=500,
		word_embedding_init=Normal(),
		context_embedding_init=Normal(),
	):

		self.input_var = input_var
		self.batch_size = batch_size
		self.vocabulary_size = vocabulary_size
		self.num_embedding_dimensions = num_embedding_dimensions
		self._embed = None

		# Every row (example) in a batch consists of a query word and a 
		# context word.  We need to learn separate embeddings for each,
		# and so, the input will be separated into all queries and all 
		# contexts, and handled separately.
		self.query_input = input_var[:,0]
		self.context_input = input_var[:,1]

		# Make separate input layers for query and context words
		#print 'using batch size:', batch_size
		self.l_in_query = lasagne.layers.InputLayer(
			shape=(batch_size,), input_var=self.query_input
		)
		self.l_in_context = lasagne.layers.InputLayer(
			shape=(batch_size,), input_var=self.context_input
		)

		# Make separate embedding layers for query and context words
		self.l_embed_query = lasagne.layers.EmbeddingLayer(
			incoming=self.l_in_query,
			input_size=vocabulary_size, 
			output_size=num_embedding_dimensions, 
			W=word_embedding_init
		)
		self.l_embed_context = lasagne.layers.EmbeddingLayer(
			incoming=self.l_in_context,
			input_size=vocabulary_size, 
			output_size=num_embedding_dimensions, 
			W=context_embedding_init
		)

		# Obtain the embedded query words and context words
		self.query_embedding = get_output(self.l_embed_query)
		self.context_embedding = get_output(self.l_embed_context)

		# We now combine the query and context embeddings, taking the 
		# dot product between corresponding queries and contexts.  
		# We can express this as matrix multiplication: if we multiply
		# the query embeddings (a matrix) with the transposed context
		# embeddings (also a matrix), the elements of the result give
		# the dot prodcuts of all context embeddings with all query
		# embeddings.  We only want the dot products for the queries
		# and contexts that came from the same example (not the dot products
		# formed by all pairs), which we can obtain by selecting the 
		# diagonal from the result.  Hopefully theano's optimizations are
		# smart enough to only calculate the needed dot products.
		self.match_scores = row_dot(
			self.query_embedding,
			self.context_embedding
		)
		#self.match_scores = T.dot(
		#	self.query_embedding, self.context_embedding.T
		#).diagonal()

		# Finally apply the sigmoid activation function
		self.output = sigmoid(self.match_scores)


	def embed(self, token_id):
		'''
		Return the vectorspace embedding for the token.

		INPUTS
		* token [int] or [iterable : int]: Integer representation of a 
			single token (e.g. a word), or a list of tokens.
		'''
		if self._embed is None:
			self._compile_embed()

		# This function is overloaded to accept a single token_id or 
		# a list thereof.  Resolve that here.
		if isinstance(token_id, int):
			token_ids = [token_id]
		else:
			token_ids = token_id

		return self._embed(token_ids)

			

	def _compile_embed(self):
		'''
		Compiles the embedding function.  This is a separate theano 
		function from that used during training of the network, but it
		produces exactly the same embeddings.  After being compiled
		once (usually the first time self.embed() is called), it will 
		remain up-to-date, supplying the correct embeddings even if
		more training occurs.  This is because it references the same
		theano shared variables for its embedding parameters as the
		embedding layer used in training.

		(No Inputs)
		OUTPUTS
		* [None]
		'''
		input_tokens = T.ivector('token')
		l_in = lasagne.layers.InputLayer(
			shape=(None,), input_var=input_tokens
		)
		l_emb = lasagne.layers.EmbeddingLayer(
			incoming=l_in,
			input_size=self.vocabulary_size, 
			output_size=self.num_embedding_dimensions, 
			W=self.l_embed_query.W
		)
		embedding = get_output(l_emb)
		self._embed = function([input_tokens], embedding)


	def get_param_values(self):
		'''
		returns a list of the trainable parameters *values*, as 
		np.ndarray's.
		'''
		return (
			get_all_param_values(self.l_embed_query, trainable=True) +
			get_all_param_values(self.l_embed_context, trainable=True)
		)


	def get_params(self):
		'''returns a list of the trainable parameters, that is, the query
		and context embeddings.  (similar to layer.get_all_params.)'''
		return (
			get_all_params(self.l_embed_query, trainable=True) +
			get_all_params(self.l_embed_context, trainable=True)
		)


	def get_output(self):
		'''returns the symbolic output.  (similar to layer.get_output.)'''
		return self.output


	def save(self, filename):
		W, C = self.get_param_values()
		np.savez(filename, W=W, C=C)


	def load(self, filename):
		npfile = np.load(filename)
		W_loaded, C_loaded = npfile['W'], npfile['C']

		# Get the theano shared variables that are used to hold the 
		# parameters, and fill them with the values just loaded
		W_shared, C_shared = self.get_params()
		W_shared.set_value(W_loaded)
		C_shared.set_value(C_loaded)




'''
Calculates the maximum likelihood probability of single words, as equal
to their frequency within a corpus, and enables efficiently sampling
from the distribution.

Assumes that the words have been converted into ids that were assigned
by auto-incrementing from 0 

(The consequence if that is not the case isn't huge, it just means memory
would be wasted)
'''

import numpy as np
import gzip


class CounterSamplerException(Exception):
	pass


class CounterSampler(object):

	def __init__(self, counts=[]):
		'''
		Create a CounterSampler.  Most common usage is to provide no
		arguments, creating an empty CounterSampler, because CounterSampler 
		provides functions to accumulate counts from observations.

		counts:	list of counts.  The index of the count in the counts list
			serves as the id for the outcome having that many counts.
		'''

		# Validation
		if not all([isinstance(c, int) for c in counts]):
			raise ValueError('Counts must be an iterable of integers')
		self.counts = list(counts)
		self.sampler_ready = False


	def add(self, idx):
		self.sampler_ready = False
		if len(self.counts) <= idx:
			add_zeros = [0] * (idx +1 - len(self.counts))
			self.counts.extend(add_zeros)

		self.counts[idx] += 1


	def update(self, idxs):
		for idx in idxs:
			self.add(idx)


	def get_sampler(self):
		if len(self.counts) < 1:
			raise CounterSamplerException(
				'Cannot sample if no counts have been made'
			)

		if not self.sampler_ready:
			self._sampler = MultinomialSampler(self.counts)
			self.sampler_ready = True

		return self._sampler


	def get_total_counts(self):
		return self.get_sampler().total
	def __len__(self):
		return self.get_sampler().total


	def sample(self, shape=()):
		'''
		Draw a sample according to the counter_sampler probability
		'''

		return self.get_sampler().sample(shape)


	def save(self, filename):
		if filename.endswith('.gz'):
			f = gzip.open(filename, 'w')
		else:
			f = open(filename, 'w')	
		
		for c in self.counts:
			f.write('%d\n' % c)


	def load(self, filename):
		if filename.endswith('.gz'):
			f = gzip.open(filename)
		else:
			f = open(filename)

		self.counts = [int(c) for c in f.readlines()]


	def get_frequency(self, idx):
		'''
		Return the number of observations for outcome idx.
		'''
		return self.counts[idx]

	def get_probability(self, token_id):
		'''
		Return the probability associated to token_id.
		'''
		# Delegate to the underlying MultinomialSampler
		return self.get_sampler().get_probability(token_id)



class MultinomialSampler(object):

	def __init__(self, scores):
		if len(scores) < 1:
			raise ValueError('The scores list must have length >= 1')
		self.scores = scores
		self.total = float(sum(scores))
		self.K = len(scores)
		self.setup()


	def get_probability(self, k):
		'''
		Get the actual probability associated to outcome k
		'''
		return self.orig_prob_mass[k]


	def setup(self):
		self.orig_prob_mass = np.zeros(self.K)
		self.mixture_prob_mass = np.zeros(self.K)
		self.mass_reasignments = np.zeros(self.K, dtype=np.int64)
 
		# Sort the data into the outcomes with probabilities
		# that are larger and smaller than 1/K.
		smaller = []
		larger  = []
		for k, score in enumerate(self.scores):
			self.orig_prob_mass[k] = score / self.total
			self.mixture_prob_mass[k] = (self.K*score) / self.total
			if self.mixture_prob_mass[k] < 1.0:
				smaller.append(k)
			else:
				larger.append(k)
 
		# We will have k different slots. Each slot represents 1/K
		# prbability mass, and to each we allocate all of the probability
		# mass from a "small" outcome, plus some probability mass from
		# a "large" outcome (enough to equal the total 1/K).
		# We keep track of the remaining mass left for the larger outcome,
		# allocating the remainder to another slot later.
		# The result is that the kth has some mass allocated to outcome
		# k, and some allocated to another outcome, recorded in J[k].
		# q[k] keeps track of how much mass belongs to outcome k, and 
		# how much belongs to outcome J[k].
		while len(smaller) > 0 and len(larger) > 0:
			small_idx = smaller.pop()
			large_idx = larger.pop()
 
			self.mass_reasignments[small_idx] = large_idx
			self.mixture_prob_mass[large_idx] = (
				self.mixture_prob_mass[large_idx] -
				(1.0 - self.mixture_prob_mass[small_idx])
			)
 
			if self.mixture_prob_mass[large_idx] < 1.0:
				smaller.append(large_idx)
			else:
				larger.append(large_idx)
 
		return self.mass_reasignments, self.mixture_prob_mass


	def sample(self, shape=()):

		if len(shape) < 1:
			return self._sample()
		else:
			this_dim = shape[0]
			recurse_shape = shape[1:]
			return np.array(
				[self.sample(recurse_shape) for i in range(this_dim)]
			, dtype='int64')


	def _sample(self):

		# Draw from the overall uniform mixture.
		k = np.int64(int(np.floor(np.random.rand()*self.K)))
	 
		# Draw from the binary mixture, either keeping the
		# small one, or choosing the associated larger one.
		if np.random.rand() < self.mixture_prob_mass[k]:
			return k
		else:
			return self.mass_reasignments[k]




import time
from multiprocessing import Queue, Process, Pipe
from Queue import Empty
from corpus_reader import CorpusReader, default_parse
from counter_sampler import MultinomialSampler
from token_map import UNK
from unigram_dictionary import UnigramDictionary, WARN, SILENT
import numpy as np
import gzip
import os


class MinibatchGenerator(object):

	NOT_DONE = 0
	DONE = 1

	def __init__(
		self,
		files=[],
		directories=[],
		skip=[],
		unigram_dictionary=None,
		noise_ratio=15,
		kernel=[1,2,3,4,5,5,4,3,2,1],
		t = 1.0e-5,
		batch_size = 1000,
		parse=default_parse,
		verbose=True
	):

		# Get a corpus reader
		self.corpus_reader = CorpusReader(
			files=files, directories=directories, skip=skip, parse=parse,
			verbose=verbose
		)

		# Load the unigram_dictionary
		if unigram_dictionary is not None:
			self.unigram_dictionary = unigram_dictionary
		else:
			self.unigram_dictionary = UnigramDictionary()

		self.noise_ratio = noise_ratio
		self.kernel = kernel
		self.t = t
		self.batch_size = batch_size

		# Validate the kernel.  It should reflect the relative 
		# frequencies of choosing tokens from a window of +/- K tokens
		# relative to a query token.  So it must have an even number of
		# entries
		if not len(self.kernel) % 2 == 0:
			raise ValueError(
				'kernel should reflect the relative frequencies of '
				'selecting a context token within +/- K of the query '
				'token, and so should have an equal number of entries '
				'defining frequencies to the left and right of the query '
				'token, and so should have an even number of entries.'
			)


	def get_vocab_size(self):
		'''
		Get the size of the vocabulary.  Only makes sense to call this
		after MinibatchGenerator.prepare() has been called, or if an
		existing (pre-filled) UnigramDictionary was loaded, since otherwise 
		it would just return 0.
		'''
		# Delegate to the underlying UnigramDictionary
		return len(self.unigram_dictionary)


	def load(self, directory):
		'''
		Load the unigram_dictionary whose files are stored in <directory>.
		'''
		# Delegate to the underlying UnigramDictionary
		self.unigram_dictionary.load(directory)

	
	def save(self, directory):
		'''
		Save the unigram_dictionary to <directory>.
		'''
		# Delegate to the underlying UnigramDictionary
		self.unigram_dictionary.save(directory)


	def check_access(self, savedir):

		savedir = os.path.abspath(savedir)
		path, dirname = os.path.split(savedir)

		# Make sure that the directory we want exists (make it if not)
		if not os.path.isdir(path):
			raise IOError('%s is not a directory or does not exist' % path)
		if not os.path.exists(savedir):
			os.mkdir(os.path)
		elif os.path.isfile(savedir):
			raise IOError('%s is a file. % savedir')

		# Make sure we can write to the file
		f = open(os.path.join(savedir, '.__test-w2v-access'), 'w')
		f.write('test')
		f.close
		os.remove(os.path.join(savedir, '.__test-w2v-access'))


	def prepare(self, savedir=None, min_frequency=None):
		'''
		Iterate over the entire corpus in order to build a 
		UnigramDictionary.  We need this because we need to sample
		from the unigram distribution in producing minibatches.
		Optionally prune all tokens that occur fewer than min_frequency
		times from dictionary.  Use min_frequency=None (the default) to
		specify no pruning.  Optionally save the dictionary to savedir 
		(this is done after pruning if pruning is requested).
		'''

		# Before doing anything, if we were requested to save the 
		# dictionary, make sure we'll be able to do that (fail fast)
		if savedir is not None:
			self.check_access(savedir)

		# Read through the corpus, building the UnigramDictionary
		for line in self.corpus_reader.read_no_q():
			self.unigram_dictionary.update(line)

		# Prune the dictionary, if requested to do so.
		if min_frequency is not None:
			self.unigram_dictionary.prune(min_frequency)

		# Save the dictionary, if requested to do so.
		if savedir is not None:
			self.save(savedir)


	def prune(self, min_frequency=5):
		'''
		Exposes the prune function for the underlying UnigramDictionary
		'''
		self.unigram_dictionary.prune(min_frequency)
  

	def __iter__(self):

		# Once iter is called, a subprocess will be started which
		# begins generating minibatches.  These accumulate in a queue
		# and iteration pulls from that queue.  That way, iteration
		# can begin as soon as the first minibatch is prepared, and 
		# later minibatches are prepared in the background while earlier
		# minibatches are used.  The idea is that this will keep the 
		# CPU(s) busy while training occurs on the GPU

		self.minibatches = Queue()
		self.recv_pipe, send_pipe = Pipe()

		# We'll fork a process to assemble minibatches, and return 
		# immediatetely so that minibatches can be used as they are 
		# constructed.
		#
		# Because we assemble the batches within a forked process, it's 
		# access to randomness doesn't alter the state of the parent's 
		# random number generator.  Multiple calls to this function
		# would produce the same set of random samples, which is not
		# desired.  We make a call to the numpy random number generator
		# to advance the parent's random number generator's state to avoid
		# this problem:
		np.random.uniform()

		minibatch_preparation = Process(
			target=self.enqueue_minibatches,
			args=(self.minibatches, send_pipe)
		)
		minibatch_preparation.start()

		return self


	def init_batch(self):
		# Initialize np.array's to store the minibatch data.  We know
		# how big the batch is ahead of time.  Initialize by filling
		# the arrays with UNK tokens.  Doing this means that, at the end
		# of the corpus, when we don't necessarily have a full minibatch,
		# the final minibatch is padded with UNK tokens in order to be
		# of the desired shape.  This has no effect on training, because
		# we don't care about the embedding of the UNK token
		signal_batch = np.full(
			(self.batch_size, 2), UNK, dtype='int32'
		)
		noise_batch = np.full(
			(self.batch_size * self.noise_ratio, 2), UNK, dtype='int32'
		)
		return signal_batch, noise_batch


	def generate(self):

		chooser = TokenChooser(K=len(self.kernel)/2, kernel=self.kernel)
		signal_batch, noise_batch = self.init_batch()

		# i keeps track of position in the signal batch
		i = -1
		for line in self.corpus_reader.read_no_q():

			# Isolated tokens (e.g. one-word sentences) have no context
			# and can't be used for training.
			if len(line) < 2:
				continue

			token_ids = self.unigram_dictionary.get_ids(line)

			# We'll now generate generate signal examples and noise
			# examples for training
			for query_token_pos, query_token_id in enumerate(token_ids):

				# Possibly discard the token
				if self.do_discard(query_token_id):
					continue

				# Increment position within the batch
				i += 1

				# Sample a token from the context
				context_token_pos = chooser.choose_token(
					query_token_pos, len(token_ids)
				)
				context_token_id = token_ids[context_token_pos]
				signal_batch[i, :] = [query_token_id, context_token_id]

				# Sample tokens from the noise
				noise_context_ids = self.unigram_dictionary.sample(
					(self.noise_ratio,))

				# Figure out the position within the noise batch
				j = i*self.noise_ratio

				# block-assign the noise samples to the noise batch array
				noise_batch[j:j+self.noise_ratio, :] = [
					[query_token_id, noise_context_id]
					for noise_context_id in noise_context_ids
				]

				# Once we've finished assembling a minibatch, enqueue it
				# and start assemblin a new minibatch
				if i == self.batch_size - 1:
					yield (signal_batch, noise_batch)
					signal_batch, noise_batch = self.init_batch()
					i = -1

		# Normally we'll have a partially filled minibatch after processing
		# the corpus.  The elements in the batch that weren't overwritten
		# contain UNK tokens, which act as padding.  Enqueue the partial
		# minibatch.
		if i >= 0:
			yield (signal_batch, noise_batch)


	def get_minibatches(self):
		'''
		Reads through the entire corpus, generating all of the minibatches
		up front, storing them in memory as a list.  Returns the list of
		minibatches.
		'''
		minibatches = []
		for signal_batch, noise_batch in self.generate():
			minibatches.append((signal_batch, noise_batch))

		return minibatches


	def enqueue_minibatches(self, minibatch_queue, send_pipe):

		'''
		Reads through the minibatches, placing them on a queue as they
		are ready.  This usually shouldn't be called directly, but 
		is used when the MinibatchGenerator is treated as an iterator, e.g.:

			for signal, noise in my_minibatch_generator:
				do_something_with(signal, noise)

		It causes the minibatches to be prepared in a separate process
		using this function, placing them on a queue, while a generator
		construct pulls them off the queue as the client process requests
		them.  This keeps minibatch preparation running in the background
		while the client process is busy processing previously yielded 
		minibatches.
		'''

		# Continuously iterate through the dataset, enqueing each
		# minibatch.  The consumer will process minibatches from
		# the queue at it's own pace.
		for signal_batch, noise_batch in self.generate():
			minibatch_queue.put((signal_batch, noise_batch))

		# Notify parent process that iteration through the corpus is
		# complete (so it doesn't need to wait for more minibatches)
		send_pipe.send(self.DONE)


	def do_discard(self, token_id):
		'''
		This function helps with downsampling of very common words.
		Returns true when the token should be discarded as a query word
		'''
		probability = self.unigram_dictionary.get_probability(token_id)
		discard_probability = 1 - np.sqrt(self.t/probability)
		do_discard = np.random.uniform() < discard_probability

		#if do_discard:
		#	print 'discarding', self.unigram_dictionary.get_token(token_id)

		return do_discard


	def next(self):
		status = self.NOT_DONE
		while status == self.NOT_DONE:
			try:
				return self.minibatches.get(timeout=0.1)
			except Empty:
				if self.recv_pipe.poll():
					status = self.recv_pipe.recv()

		raise StopIteration

				

class TokenChooser(object):

	'''
	This choses which context token should be taken given a window
	of +/- K around a query token
	'''

	def __init__(self, K, kernel):
		if not len(kernel) == 2*K:
			raise ValueError(
				'`kernel` must have 2*K entries, one for '
				'each of the elements within the windows of +/- K tokens.'
			)

		self.K = K
		self.kernel = kernel
		self.samplers = {}
		self.indices = range(-K, 0) + range(1, K+1)


	def choose_token(self, idx, length):
		'''
		Randomly choosea  token according to the kernel supplied
		in the constructor.  Note that when sampling the context near
		the beginning of a sentence, the left part of the context window
		will be truncated.  Similarly, sampling context near the end of
		a sentence leads to truncation of the right part of the context 
		window.  Short sentences lead to truncation on both sides.

		To ensure that samples are returned within the possibly truncated
		window, two values define the actual extent of the context to be
		sampled:

		`idx`: index of the query word within the context.  E.g. if the
			valid context is constrained to a sentence, and the query word
			is the 3rd token in the sentence, idx should be 2 (because
			of 0-based indexing)

		`length`: length of the the context, E.g. If context is 
			constrained to a sentence, and sentence is 7 tokens long,
			length should be 7.
		'''

		# If the token is near the edges of the context, then the 
		# sampling kernel will be truncated (we can't sample before the 
		# firs word in the sentence, or after the last word).
		# Determine the slice indices that define the truncated kernel.
		negative_idx = length - idx
		start = max(0, self.K - idx)
		stop = min(2*self.K, self.K + negative_idx - 1)	

		# We make a separate multinomial sampler for each different 
		# truncation of the kernel, because they each define a different 
		# set of sampling probabilityes.  If we don't have a sampler for 
		# this particular kernel shape, make one.
		if not (start, stop) in self.samplers:

			trunc_probabilities = self.kernel[start:stop]
			self.samplers[start,stop] = (
				MultinomialSampler(trunc_probabilities)
			)

		# Sample from the multinomial sampler for the context of this shape
		outcome_idx = self.samplers[start,stop].sample()

		# Map this into the +/- indexing relative to the query word
		relative_idx = self.indices[outcome_idx + start]

		# And then map this into absolute indexing
		result_idx = relative_idx + idx

		return result_idx





import os 
from multiprocessing import Process, Pipe, Queue
from Queue import Empty
import sys


def default_parse(filename):
	tokenized_sentences = []
	for line in open(filename):
		tokenized_sentences.append(line.strip().split())
	return tokenized_sentences



class CorpusReader(object):

	NOT_DONE = 0
	DONE = 1

	def __init__(
		self, 
		files=[],
		directories=[],
		skip=[],
		parse=default_parse,
		verbose=True
	):
		self.files = files
		self.directories = directories
		self.skip = skip
		self.parse = parse
		self.verbose = verbose


	def read_no_q(
		self
	):
		'''
		Iterates through the files and directories given in the constructor
		and parses out a list of sentences, where sentences are encoded
		as a list of tokens (i.e. a list of lists of tokens).
		Parsing the files is deligated to a parser function, which can
		be customized.

		Lines are loaded into a queue so that reading can be done in the
		background.
		'''

		# Process all the files listed in files, unles they match an
		# entry in skip
		#print 'starting reading'
		if self.files is not None:
			for filename in self.files:
				filename = os.path.abspath(filename)

				# Skip files if they match a regex in skip
				if any([s.search(filename) for s in self.skip]):
					continue

				if self.verbose:
					print 'processing', filename
				for line in self.parse(filename):
					yield(line)

		# Process all the files listed in each directory, unless they
		# match an entry in skip
		if self.directories is not None:
			for dirname in self.directories:
				dirname = os.path.abspath(dirname)

				# Skip directories if they match a regex in skip
				if any([s.search(dirname) for s in self.skip]):
					continue

				for filename in os.listdir(dirname):
					filename = os.path.join(dirname, filename)

					# Only process the *files* under the given directories
					if not os.path.isfile(filename):
						continue

					# Skip files if they match a regex in skip
					if any([s.search(filename) for s in self.skip]):
						continue

					if self.verbose:
						print 'processing', filename
					for line in self.parse(filename):
						yield line



	def read(self):

		queue = Queue()
		pipe1, pipe2 = Pipe()

		# We do the reading in a separate process, that way, if the
		# consumer is busy processing the read items, we keep loading
		# the corpus in the background
		reading_process = Process(
			target=self._read, 
			args=(self.files, self.directories, self.skip, queue, pipe2)
		)
		reading_process.start()

		state = self.NOT_DONE

		while state == self.NOT_DONE:

			try:
				yield queue.get(timeout=0.1)

			except Empty:
				if pipe1.poll():
					state = pipe1.recv()


	def _read(
		self,
		files=[],
		directories=[],
		skip=[],
		queue=None,
		pipe=None
	):
		'''
		Iterates through the files and directories given in the constructor
		and parses out a list of sentences, where sentences are encoded
		as a list of tokens (i.e. a list of lists of tokens).
		Parsing the files is deligated to a parser function, which can
		be customized.

		Lines are loaded into a queue so that reading can be done in the
		background.
		'''

		# Process all the files listed in files, unles they match an
		# entry in skip
		if files is not None:
			for filename in files:
				filename = os.path.abspath(filename)

				# Skip files if they match a regex in skip
				if any([s.search(filename) for s in skip]):
					continue

				for line in self.parse(filename):
					queue.put(line)

		# Process all the files listed in each directory, unless they
		# match an entry in skip
		if directories is not None:
			for dirname in directories:
				dirname = os.path.abspath(dirname)

				# Skip directories if they match a regex in skip
				if any([s.search(dirname) for s in skip]):
					continue

				for filename in os.listdir(dirname):
					filename = os.path.join(dirname, filename)

					# Only process the *files* under the given directories
					if not os.path.isfile(filename):
						continue

					# Skip files if they match a regex in skip
					if any([s.search(filename) for s in skip]):
						continue

					for line in self.parse(filename):
						queue.put(line)

		# Notify the parent process that you're done
		if pipe is not None:
			pipe.send(self.DONE)






class Word2Vec(object):

	def __init__(
		self,
		batch_size,
		vocabulary_size=100000,
		num_embedding_dimensions=500,
		word_embedding_init=Normal(), 
		context_embedding_init=Normal(),
		learning_rate=0.1,
		momentum=0.9,
		verbose=True
	):

		# Register all the arguments for easy inspection and access
		self.batch_size = batch_size
		self.vocabulary_size = vocabulary_size
		self.num_embedding_dimensions = num_embedding_dimensions
		self.word_embedding_init = word_embedding_init
		self.context_embedding_init = context_embedding_init
		self.learning_rate = learning_rate
		self.momentum = momentum
		self.verbose = verbose

		# The input that we pass through the Word2VecEmbedder will be
		# the concatenation of the signal and noise examples
		self.positive_input = T.imatrix('positive_input')
		self.negative_input = T.imatrix('negative_input')
		embedder_input = T.concatenate([
			self.positive_input, self.negative_input
		])

		# Make a Word2VecEmbedder
		self.embedder = Word2VecEmbedder(
			embedder_input,
			batch_size,
			vocabulary_size=vocabulary_size,
			num_embedding_dimensions=num_embedding_dimensions,
			word_embedding_init=word_embedding_init,
			context_embedding_init=context_embedding_init
		)

		# Split the output back up into its positive and negative parts
		embedder_output = self.embedder.get_output()

		self.positive_output = embedder_output[
			:self.positive_input.shape[0]
		]
		self.negative_output = embedder_output[
			self.positive_input.shape[0]:
		]

		# Construct the loss based on Noise Contrastive Estimation
		self.loss = noise_contrast(
			self.positive_output, self.negative_output
		)

		# Get the parameter updates using stochastic gradient descent with
		# nesterov momentum.  TODO: make the updates configurable
		self.updates = lasagne.updates.nesterov_momentum(
			self.loss,
			self.embedder.get_params(), 
			learning_rate,
			momentum
		)

		# That's everything we need to compile a training function.
		# The training fuction will be defined separately, and will 
		# wrap the compiled theano function, which will be compiled when
		# it is first called.


	def get_param_values(self):
		'''
		Returns the values of the embedding parameters, from the 
		underlying embedder.
		'''
		return self.embedder.get_param_values()

	def get_params(self):
		'''
		Returns the embedding parameters, of the underlying embedders.
		'''
		return self.embedder.get_params()


	def train_from_corpus(
		self, files=[], directories=[], skip=[],
		loaddir=None, savedir=None
	):
		'''
		Takes an iterable of paths that point to corpus files / folders.
		Reads those files, and:
		0) Reads sentence-split, tokenized documents in a way that 
			manages memory and avoid's blocking downstream steps during
			I/O requests.
			- a reading process stays ahead of the downstream process in
				loading files into memory
			- a customizable parser is responsible for understanding the
				file format
			- `files` and `directories` can be strings, or iterables of
				strings, including custom generators that walk the 
				filesystem
		1) Makes a token_map which encodes tokens as integers
			- the token_map class stores a two-way mapping between 
				token types (i.e. unique words) and integers
			- forward mapping from tokens to integers uses a token_map
				with keys as tokens and ints as values
			- revrese mapping from integers to tokens uses a list of tokens
		2) Makes a counter_sampler model which enables sampling from the counter_sampler
			model
			- keeps count of how many times a given tokens were seen
				(using a list whose values are counts and whose 
					components are the token ids)
			- enables rapid sampling from the counter_sampler distribution. Using
				a binary tree structure.  The first time sampling is 
				attempted after an update has been made to the counts,
				the tree is updated / generated.
			- a full pass through the corpus is needed to create the counter_sampler
				model.
		3) Generates batches of signal and noise examples to train the
			embeddings


		If `savedir` is specified, then three files will be saved into
		the directory specified by savedir (savedir will be created if 
		it doesn't exist, as long as other dirs in its path already exist)
		1) savedir/token_map.gz -- stores the token_map mapping
		2) savedir/counter_sampler.gz -- stores the counter_sampler frequencies.  This 
			means that future training using different sampling will
			not need to re-count the counter_sampler frequencies!
		3) savedir/embeddings.gz -- stores the embedding parameters for
			both query-embeddings (which is the main word embedding of use
			in other applications) as well as the context-embedding (not
			usually needed for other applications, but kept for 
			completeness)

		if `loaddir` is specified, the token_map and counter_sampler saved in
		loaddir/token_map.gz and loaddir/counter_sampler.gz will be loaded.
		This means that the counter_sampler frequencies (and token_map) don't 
		need to be made before training.
		'''

		# If a savedir is given, check that it is valid before proceeding
		# that way we fail fast if there's an IO issue
		if savedir is not None:

			# if savedir exists, just make sure it's not a file
			if os.path.exsits(savedir):
				if os.path.isfile(savedir):
					raise IOError(
						'Path supplied as `savedir` is a file: %s'
						% savedir
					)

			# if savedir doesn't exist, try to make it
			else:
				os.mkdir(savedir)

		# Create a token_map.  We'll pass it into the minibatch generator,
		# but we want a reference to it in the Word2Vec object too, 
		# since it also needs the token_map
		token_map = TokenMap()

		# Make a minibatch generator
		minibatch_generator = Word2VecMinibatcher(
			files, directories, skip, token_map
		)

		# Run the minibatch generator over the corpus to collect counter_sampler
		# statistics and to fill out the token_map
		minibatch_generator.prepare_counter_sampler()

		# We now have a full token_map and the counter_sampler statistics.
		# Save them if savedir was specified
		if savedir is not None:
			token_map.save(os.path.join(savedir, 'token_map.gz'))
			minibatch_generator.save_counter_sampler(
				os.path.join(savedir, 'counter_sampler.gz')
			)

		# Here is where the training actually happens
		for positive_input, negative_input in minibatch_generator:
			self.train(positive_input, negative_input)

		# Save the learned embedding (if savedir was specified)
		if savedir is not None:
			self.embedder.save(os.path.join(savedir, 'embedding.npz'))


	def save(self, savedir):
		self.token_map.save(os.path.join(savedir, 'token_map.gz'))
		self.embedder.save(os.path.join(savedir, 'embedding.npz'))


	def train(self, positive_input, negative_input):
		'''
		Runs the compiled theano training function, using the positive
		query-context pairs contained in positive_input and the negative
		query_context pairs contained in negative_input.  Both are expected
		to be numpy int32-type arrays, with shape (batch_size, 2), each
		example is a row with two elements: the query-word and the
		context-word, coded as integers.
  
		If the theano training function hasn't been compiled yet (which
		is true the first time this is called, it compiles it.
		'''
		if not hasattr(self, 'train_'):
			if self.verbose:
				print 'Compiling training function on first call.'
			self.train_ = function(
				inputs=[self.positive_input, self.negative_input], 
				outputs=self.loss,
				updates=self.updates
			)

		return self.train_(positive_input, negative_input)


